unit NZCSSpellCheck;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,stdctrls;

Const
   MaxWordLength = 30;
   SpellRegPath = 'Software\NZCS\Spelling';
   WordDelimiters: set of Char = [#0..#255] - ['a'..'z','A'..'Z','1'..'9','0'];

type
  TDictWords = string[MaxWordLength];
  TDialogOptions = (scIgnore, scIgnoreAll, scCancel, scChange, scChangeAll, scAdd, scRestart);
  TSpellingOptions = (spIgnoreNumbers, spIgnoreAllUpperCase, spIgnoreInternet, spIgnoreFiles,spIgnoreBrackets,
                      spMainDicOnly, spDuplicateSpaces, spDuplicateWords, spAlwaysSuggest);
  TSpellingOptionsSet = set of TSpellingOptions;


  TNZCSSpellCheck = class(TComponent)
  private
    FileOpen: Boolean;
    FSoundexLength: Integer;
    FSpellingOptions: TSpellingOptionsSet;
    fdictfile: file of TDictWords;
    FDictionary: TFileName;
    FCustomDictionary: TFileName;
    FIgnoreAll: TStringList;
    FCustomDicList: TStringList;
    FWordCount: Integer;
    procedure SetDictionary(Value: TFileName);
    procedure SetCustomDictionary(Value: TFileName);
    procedure SetIgnoreAll(Value: TStringList);
    procedure CloseDict;

    function StripWord(const Key: String): String;
    function RemoveDupSpace(Memo: TCustomMemo; FirstDup: Integer): Boolean;
    function RepeativeWord(Memo: TCustomMemo; LastWordStart, LastWordEnd, StartWord, EndWord: Integer): Boolean;
    function MisSpeltWord(Memo: TCustomMemo; Word: String; StartWord, EndWord: Integer): boolean;
    function SearchMemo(Memo: TCustomMemo;
      const SearchString: String): Boolean;
  protected
    DictIndex: array[1..26] of Integer;
  public
     Cancelled: Boolean;
     function FindCustomDic: String;
     procedure LoadSpellOptions;
     procedure SaveSpellOptions;
     function SetCase(WrongWord, Word: String): String;
     procedure CustomDictAdd(Key: String);
     Procedure ClearIgnoreAll;
     procedure CheckMemo(Memo: TCustomMemo);
     Procedure IgnoreAllAdd(Key: String);
     procedure CompileFile(const Fn: array of String);
     function GetSuggestions(Key: String; Suggestions: TStringList): Boolean;
     function FindByIndex(Idx: Integer): TDictWords;
     procedure OpenFile;
     function Soundex(Str: string): String;
     constructor Create(AOwner: TComponent); override;
     destructor Destroy; override;
     function CheckWord(key:string):boolean;
     Function WordAtCursor(Memo: TMemo): String;
     //procedure CheckText(Text: TStringList);
     property WordCount: Integer read FWordCount write FWordCount;
  published
    property Dictionary: TFileName read FDictionary write SetDictionary;
    property CustomDictionary: TFileName read FCustomDictionary write SetCustomDictionary;
    property SoundexLength: Integer read FSoundexLength write FSoundexLength default 6;
    property SpellingOptions: TSpellingOptionsSet read FSpellingOptions write FSpellingOptions
             default [spIgnoreNumbers, spIgnoreAllUpperCase, spIgnoreInternet, spIgnoreFiles, spIgnoreBrackets,
                      spDuplicateSpaces, spDuplicateWords, spAlwaysSuggest];
  end;

procedure Register;

function StripDelimiter(const Delimiters, xString: STRING): string;

implementation

uses FileCtrl, spelldlg, NZCSGrammer, Registry;

procedure Register;
begin
  RegisterComponents('Standard', [TNZCSSpellCheck]);
end;

// Utility //
function GetChar(Value: String; index: smallint): Char;
begin
   if Length(Value) < index then result := #0 else Result := Value[index];
end;

function AsChar(Value: String): Char;
begin
  result := GetChar(Value,1);
end;

Function Same(xString1, xString2: String): Boolean;
begin
   result := (CompareText(xString1, xString2) = 0);
end;

function PropperCase (strInt: STRING): string;
var
   strOut: string;
   iPos: Integer;
begin
   strInt := LowerCase(strInt);
   strOut := UpperCase(Copy(strInt, 1, 1));
   for iPos := 2 to Length(strInt) do begin
      if ((Copy(strInt, iPos - 1, 1) = ' ') and
      (UpperCase(copy(strInt, iPos, 1)) <> copy(strInt, iPos, 1)))
      then
         strOut := strOut + UpperCase(copy(strInt, iPos, 1))
      else
         strOut := strOut + Copy(strInt, iPos, 1);
   end;
   result :=  strOut;
end;

function AppPath: string; {Appliction Exe Path}
var
   TmpPath: string;
begin
   TmpPath := ExtractFilePath(Application.Exename);
   if (TmpPath <> '') and (TmpPath[Length(TmpPath)] <> '\') then
      TmpPath := TmpPath + '\';
   Result := TmpPath;
end;

function StripDelimiter(const Delimiters, xString: STRING): string;
var
   i: Integer;
begin
   Result := xString;
   i := 1;
   While i <= length(Result) do begin
      If IsDelimiter(Delimiters,Result,i) then
         Delete(Result,i,1)
      else
         inc(i);
   end;
end;

{ TNZCSSpellCheck }

constructor TNZCSSpellCheck.Create(AOwner: TComponent);
begin
   Inherited Create(AOwner);
   FIgnoreAll := TStringList.Create;
   Cancelled := False;
   FCustomDicList := TStringList.Create;
   FSoundExLength := 6;
   FSpellingOptions := [spIgnoreNumbers, spIgnoreAllUpperCase, spIgnoreInternet, spIgnoreFiles, spIgnoreBrackets,
                        spDuplicateSpaces, spDuplicateWords, spAlwaysSuggest];
   If not (csDesigning in ComponentState) then
      LoadSpellOptions;
end;

destructor TNZCSSpellCheck.Destroy;
begin
   CloseDict;
   FIgnoreAll.Free;
   FIgnoreAll := nil;
   FCustomDicList.Free;
   FCustomDicList := nil;
   if assigned(FrmSpellCheck) then begin
      FrmSpellCheck.Free;
      FrmSpellCheck := nil;
   end;
   if assigned(FrmSpellGrammer) then begin
      FrmSpellGrammer.Free;
      FrmSpellGrammer := nil;
   end;
   Inherited Destroy;
end;

function TNZCSSpellCheck.CheckWord(key: string): boolean;
Var
   DictWord: TDictWords;
   DIndx, StartIdx, HighIdx, Mid: Integer;
begin
   result := True;
   Key := StripWord(Trim(Key));
   If (Key = '') or (FIgnoreAll.IndexOf(Key) <> -1) then begin
      Exit;
   end;
   // Get the upper case ascii value of the first letter
   DIndx := Ord(AsChar(Key)) - 64;
   If (DIndx > 0) and (DIndx < 27) then begin
      result := False;
      OpenFile;
      // First check the custom dictionary //
      If FCustomDicList.IndexOf(Key) <> -1 then begin
         result := True;
         Exit;
      end;
      HighIdx := WordCount;
      // Get the position of the first word in our distionary starting with the words first letter //
      StartIdx := DictIndex[DIndx];
      // Unless the word starts with Z get the next letters starting position //
      If Dindx < 26 then
         HighIdx := DictIndex[DIndx+1];
      // Divide into two to make the search 50% fast if the word is at either end //
      If (StartIdx > WordCount) or (StartIdx < 0) then
         StartIdx := WordCount div 2;
      // Go up or down the file until we get to the end or find the word //
      while (StartIdx <= HighIdx) do begin
         Mid := (HighIdx+StartIdx) div 2;
         Seek(fdictfile, mid);
         Read(fdictfile, DictWord);
         If Key = DictWord then begin
            result := True;
            Break;
         end;
         if (DictWord > key) then
            HighIdx := mid-1
         else
            StartIdx := mid+1;
      end;
   end;
end;

Function TNZCSSpellCheck.SetCase(WrongWord, Word: String): String;
begin
   If WrongWord = Lowercase(WrongWord) then
      result := Lowercase(Word)
   else begin
      If WrongWord = Uppercase(WrongWord) then
         result := uppercase(Word)
      else begin
         result := PropperCase(Word);
      end;
   end;
end;


Function TNZCSSpellCheck.GetSuggestions(Key: String; Suggestions: TStringList): Boolean;
Var
   DictWord: TDictWords;
   DIndx, StartIdx, HighIdx, Mid, I: Integer;
   CSoundex, SKey: String;
   StTime: TDateTime;
   WLen: Integer;
begin
   result := False;
   SKey := UpperCase(Trim(Key));
   CSoundex := Soundex(SKey);
   DIndx := Ord(AsChar(SKey)) - 64;
   Suggestions.Clear;
   StTime := Now;
   If (DIndx > 0) and (DIndx < 27) then begin
      result := False;
      OpenFile;
      HighIdx := WordCount;
      StartIdx := DictIndex[DIndx];
      If Dindx < 26 then
         HighIdx := DictIndex[DIndx+1];
      If (StartIdx > WordCount) or (StartIdx < 0) then
         StartIdx := 0;
      while (StartIdx <= HighIdx) do begin
         Seek(fdictfile, StartIdx);
         Read(fdictfile, DictWord);
         If CSoundex = Soundex(DictWord) then begin
            result := True;
            Suggestions.Add(SetCase(Key,DictWord));
         end
         else begin
            WLen := length(DictWord);
            If (WLen >= 4) and (Copy(SKey, 1, WLen) = DictWord) then begin
               result := True;
               Suggestions.insert(0,SetCase(Key,DictWord));
            end
            else begin
               If (WLen >= 4) and (Copy(SKey, 2, MaxInt) = DictWord) then begin
                  result := True;
                  Suggestions.insert(0,SetCase(Key,DictWord));
               end
            end;
         end;
         Inc(StartIdx);
      end;
      if not (spMainDicOnly in SpellingOptions) then begin
         For i := 0 to FCustomDicList.Count - 1 do begin
            If FormatDateTime('ss:mm',StTime - Now) > '00:01' then
               break;
            DictWord := FCustomDicList[i];
            If Uppercase(Copy(DictWord,1,1))  = Copy(SKey,1,1) then begin
               If CSoundex = Soundex(DictWord) then begin
                  result := True;
                  Suggestions.Add(SetCase(Key,DictWord));
               end;
            end;
         end;
      end;
   end;
end;

procedure TNZCSSpellCheck.OpenFile;
var
   i: Integer;
   FLoadIdx: TStringList;
   OD: TOpenDialog;
   myRegistry: TRegistry;
begin
   If FileOpen then Exit;
   FileOpen := False;
   If FDictionary = '' then
      FDictionary := AppPath + 'english.dic';
   If not FileExists(FDictionary) then begin
      OD := TOpenDialog.Create(self);
      OD.Title := 'Select Dictionary';
      If not OD.Execute then begin
         Showmessage('Cannnot find dictionary');
         Abort;
      end;
      FDictionary := OD.FileName;
      myRegistry := TRegistry.Create;
      If myRegistry.OpenKey(SpellRegPath, True) then begin
         myRegistry.WriteString('Dictionary',FDictionary);
         myRegistry.CloseKey;
      end;
      myRegistry.Free;
      OD.Free;
   end;
   AssignFile(fdictfile, FDictionary);
   {Set FileMode for Read/Write};
   FileMode := 2;
   try
     Reset(fdictfile);
   except
    on EInOutError do
      Rewrite(fdictfile);
   end;
   FWordCount := FILESIZE(fdictfile);
   FLoadIdx := TStringList.Create;
   If FileExists(ChangeFileExt(FDictionary,'.idx')) then
      FLoadIdx.LoadFromFile(ChangeFileExt(FDictionary,'.idx'));
   For i := 0 to 25 do begin
      If i > FLoadIdx.Count - 1 then
         DictIndex[i+1] := (FILESIZE(fdictfile) div 26) * i
      else
         DictIndex[i+1] := StrToIntDef(Copy(FLoadIdx[i],3,MaxInt),0);
   end;
   FLoadIdx.Free;
   FCustomDicList.Clear;
   If (FCustomDictionary <> '') and FileExists(FCustomDictionary) then
      FCustomDicList.LoadFromFile(FCustomDictionary);
   FileOpen := True;
end;

Procedure TNZCSSpellCheck.CloseDict;
begin
   If FileOpen then begin
      {$i-}
      CloseFile(fdictfile);
      {$i+}
      if ioresult <> 0 then;
         FileOpen := False;
   end;
end;

Function TNZCSSpellCheck.FindByIndex(Idx: Integer): TDictWords;
begin
   result := '';
   //result.Soundex := '';
   OpenFile;
   If not FileOpen then Exit;
   If (Idx <= FWordCount) and (idx >= 0) then begin
      Seek(fdictfile, Idx);
      If not EOF(fdictfile) then
         Read(fdictfile, result);
   end;
end;

// Create a dictionary file and an index file //

{NZCSSpellCheck := TNZCSSpellCheck.Create(Application);
   NZCSSpellCheck.CompileFile(['f:\spell\Compile\Words1.lst', 'f:\spell\Compile\Words2.lst',
   'f:\spell\Compile\Words2.lst','f:\spell\Compile\Words4.lst']);
   NZCSSpellCheck.Free;}

Procedure TNZCSSpellCheck.CompileFile(const Fn: array of String);
var
   WordList: TStringList;
   TmpList: TStringList;
   DictWords: TDictWords;
   LastLetter: String[1];
   i: Integer;
begin
   CloseDict;
   If FDictionary = '' then
      FDictionary := AppPath + 'english.dic';
   AssignFile(fdictfile, FDictionary);
   {Set FileMode for Read/Write};
   FileMode := 2;
   Rewrite(fdictfile);
   FileOpen := True;
   WordList := TStringList.Create;
   TmpList := TStringList.Create;
   For i := 0 to High(Fn) do begin
      TmpList.LoadFromFile(Fn[i]);
      WordList.AddStrings(TmpList);
      TmpList.Clear;
   end;
   TmpList.Clear;
   //WordList.Sort;
   Seek(fdictfile, 0);
   For i := 0 to WordList.Count - 1 do begin
      DictWords := Uppercase(WordList[i]);
      //DictWords.Soundex := Soundex(WordList[i]);
      Seek(fdictfile, FileSize(fdictfile));
      Write(fdictfile, DictWords);
      If LastLetter <> Copy(DictWords,1,1) then begin
         LastLetter := Copy(DictWords,1,1);
         TmpList.Add(LastLetter + '=' + IntToStr(i));
      end;
   end;
   TmpList.SaveToFile(ChangeFileExt(FDictionary,'.idx'));
   CloseDict;
   WordList.Clear;
   WordList.Free;
   TmpList.Free;
end;

// Find words like //
function TNZCSSpellCheck.Soundex(Str: string): String;
var  temp : string;              {temporary adjusted target token}
        i : integer;             {index counter}
   digraph: String;

{This function inspects a two character string and encodes digraphs }
function checkdigraph(pair:string):string;
var   index : integer;
begin
   {dig string looks like: /aa=b/cc=d/ee=f}
   index := pos('/'+uppercase(pair),digraph);
   if index = 0 then
      checkdigraph := pair
   else
      checkdigraph := digraph[index+4];
end;

{This procedure checks for special cases for the first two characters}
procedure checkfirst;
begin
   i := 2;
   temp := checkdigraph(copy(Str,1,2));
   if length(temp) = 2 then         {i.e. it wasn't a digraph}
      temp := Copy(temp,1,1)  {just keep the first char}
   else
      i := 3;       {skip second char for encode}
end;

{This procedure checks for special cases for the last two characters}
procedure checklast;
var
   twochar : string[2];
begin
   twochar := copy(Str,length(Str)-1,2);
   if length(Str) > i+2 then
      temp := temp + checkdigraph(twochar)
   else
      temp := temp + twochar;
end;

 {This function returns the soundex code for a given character}
function encodechar(aChar:char):char;
begin
   case upCase(aChar) of
   'A','E','H','I','O','U','W','Y' : encodechar := '0';
   'B','F','P','V'                 : encodechar := '1';
   'C','G','J','K','Q','S','X','Z' : encodechar := '2';
   'D','T'                         : encodechar := '3';
   'L'                             : encodechar := '4';
   'M','N'                         : encodechar := '5';
   'R'                             : encodechar := '6';
   end;
end;

{This procedure sets up the temp version of the target token]}
procedure InitializeTemp;
begin;
   CheckFirst;     {checks for leading digraph; inits temp and i}
   temp := temp + copy(Str,i,length(Str)-(i+1));
   CheckLast;      {checks for trailing digraph; completes temp}
end;

{--------------------------------------------------------------------------}
{Soundexer Function Main Code                                                }
{--------------------------------------------------------------------------}
begin
   digraph  := '/GH=F/LD=D/PH=F';
   InitializeTemp;  {initialzes temp string and starting point}
   {convert temp string to soundex string}
   for i := 2 to length(temp) do
      temp[i] := encodechar(temp[i]);
   {remove doublecodes and vowels; truncate at codemax}
   Result := UpperCase(copy(temp,1,1)); {first character is always kept}
   i    := 2;
   while (length(Result) <  SoundExLength) and (i <= length(temp)) do begin
      if (temp[i] <> '0') and (temp[i] <> temp[i-1]) then
         Result := Result + temp[i];
      inc(i);
   end;
   Result := Result + '000000000000000000000000000000000000';
   Result := Copy(Result,1,SoundExLength);
end;


procedure TNZCSSpellCheck.SetDictionary(Value: TFileName);
begin
   If FDictionary <> Value then begin
      CloseDict;
      FDictionary := Value;
   end;
end;

procedure TNZCSSpellCheck.SetIgnoreAll(Value: TStringList);
begin
   FIgnoreAll.Assign(Value);
end;

//To do //
{Find Uncapitalized Start of Sentence
Find Repeated Words}

{
Ignore Numbers: e.g., 1-800-266-5626
Ignore Ordinals: e.g., 1st, 2nd, 3rd
Ignore Roman Numerals, e.g., IV, VII - Needs work but who uses Roman Numerals anyway
Ignore Parentheses in Words: e.g., sales(wo)man, shoe(s)
}

Function TNZCSSpellCheck.StripWord(const Key: String): String;
var
   TmpKey: String;

Function LeftHardTrim(S: String): String;
var
  I, L: Integer;
begin
  L := Length(S);
  I := 1;
  while (I <= L) and ((S[I] <= ' ') or (IsDelimiter(':;()-_=+=\|/.,~`[]{}!?''"<>',S,I))) do Inc(I);
  Result := Copy(S, I, Maxint);
end;

Function RightHardTrim(S: String): String;
var
  I: Integer;
begin
  I := Length(S);
  while (I > 0) and ((S[I] <= ' ') or (IsDelimiter(':;()-_=+=\|/.,~`[]{}"''!?<>',S,I))) do Dec(I);
  Result := Copy(S,1, I);
end;

Function StripComments(Str, StartChar, FinishChar: String): String;
var
   st,en: longint;
begin
   While Pos(StartChar, Str) > 0 do begin
      St := Pos(StartChar, Str);
      en := Pos(FinishChar, Str);
      If st > en then begin // If no closing brack then exit //
         Result := Str;
         break;
      end;
      Str := Copy(Str,1, st-1) + Copy(Str,en + length(FinishChar), length(str));
   end;
   Result := Str;
end;


begin
   Result := Key;
   If spIgnoreBrackets in FSpellingOptions then begin
      Result := StripComments(Result, '{', '}');
      Result := StripComments(Result, '(', ')');
      Result := StripComments(Result, '[', ']');
   end;
   Result := LeftHardTrim(RightHardTrim(Result));

   If spIgnoreNumbers in FSpellingOptions then begin
      TmpKey := StripDelimiter('0123456789-.' + DateSeparator,Key);
      If (TmpKey = '') or ((Lowercase(TmpKey) = 'st') and (key <> TmpKey)) or
         ((Lowercase(TmpKey) = 'nd') and (key <> TmpKey)) or
         ((Lowercase(TmpKey) = 'rd') and (key <> TmpKey)) or
         ((Lowercase(TmpKey) = 'th') and (key <> TmpKey)) then
         result := '';
      If (Uppercase(TmpKey) = 'II') or (Uppercase(TmpKey) = 'III') or (Uppercase(TmpKey) = 'IV') or
         (Uppercase(TmpKey) = 'V') or (Uppercase(TmpKey) = 'VI') or (Uppercase(TmpKey) = 'VII') or
         (Uppercase(TmpKey) = 'VIII') or (Uppercase(TmpKey) = 'IX') or (Uppercase(TmpKey) = 'X') or
         (Uppercase(TmpKey) = 'XI') then
         result := '';
   end;
   If (spIgnoreAllUpperCase in FSpellingOptions) and (Result <> '') and (Result = Uppercase(Result)) then begin
      result := '';
   end;
   If (spIgnoreInternet in FSpellingOptions) and (Result <> '') then begin
      If (pos('@',Result) > 0) and (pos('.',Result) > 0) then
         result := '';
      If (pos('mailto:',lowercase(Result)) > 0) or
         (pos('www.',lowercase(Result)) > 0) or
         (pos('http://.',lowercase(Result)) > 0) or
         (pos('ftp.',lowercase(Result)) > 0) or
         (pos('https://',lowercase(Result)) > 0) then
         result := '';
   end;
   If (spIgnoreFiles in FSpellingOptions) and (Result <> '') then begin
      If (Copy(Result,1,2) = '\\') or (copy(Result,2,2) = ':\') then
         result := '';
   end;
   //Result := StripLastIfDelimter(
   If result <> '' then
      Result := Uppercase(StripDelimiter('.',result));
end;

procedure TNZCSSpellCheck.SetCustomDictionary(Value: TFileName);
begin
   If FCustomDictionary <> Value then begin
      CloseDict;
      FCustomDictionary := value;
   end;
end;

procedure TNZCSSpellCheck.ClearIgnoreAll;
begin
   FIgnoreAll.Clear;
end;

procedure TNZCSSpellCheck.IgnoreAllAdd(Key: String);
begin
   If FIgnoreAll.IndexOf(Key) = -1 then begin
      FIgnoreAll.Add(Key);
   end;
end;

procedure TNZCSSpellCheck.CustomDictAdd(Key: String);
begin
   If FCustomDicList.IndexOf(key) = -1 then begin
      FCustomDicList.Add(Key);
      FCustomDicList.Sort;
      If FCustomDictionary = '' then
         CustomDictionary := FindCustomDic;
      FCustomDicList.SaveToFile(FCustomDictionary);
   end;
end;


procedure TNZCSSpellCheck.CheckMemo(Memo: TCustomMemo);
var
   i, st, lstst, lsten, lineSt,StartPos: Integer;
   txt: string;
   Word, LstWord: String;
begin
   Cancelled := False;
   With Memo do begin
      StartPos := SelStart;
      i := SelStart+1;
      st := i;
      lineSt := 1;
      Txt := Memo.Lines.Text;
      LstWord := '.,<a~*(';
      If i > 1 then begin
         While (i > 0) and (Ord(Txt[i]) > 32) do begin
            dec(i);
         end;
      end;
      While i <= Length(Memo.Text) do begin
         Txt := Memo.Text;
         While (i <= Length(txt)) and (Ord(Txt[i]) > 32) do begin
            If Txt[i] = #10 then
               lineSt := i+1;
            inc(i);
         end;
         Word := Copy(txt,st,i-st);
         If (spDuplicateWords in SpellingOptions) then begin
            If Same(Word, LstWord) then begin
               if not RepeativeWord(Memo,lstst, lsten, st,i) then begin
                  Cancelled := True;
                  Exit;
               end;
               Txt := Memo.Lines.Text;
            end;
         end;
         LstWord := Word;
         lstst := st;
         lsten := i;
         if not CheckWord(Word) then begin
            if not MisSpeltWord(Memo,Word,st,i) then begin
               Cancelled := True;
               Exit;
            end;
            LstWord := 'x~';
            Txt := Memo.Lines.Text;
            i := Memo.SelStart;
         end;
         If (spDuplicateSpaces in SpellingOptions) then begin
            if (i + 1 <= Length(txt)) and (Ord(Txt[i]) = 32) and (Ord(Txt[i+1]) = 32) then begin
               if not RemoveDupSpace(Memo,i) then begin
                  Cancelled := True;
                  Exit;
               end;
               Txt := Memo.Lines.Text;
               While (i+1 <= Length(txt)) and (Ord(Txt[i+1]) = 32) do begin
                  If Txt[i] = #10 then begin
                     lineSt := i+1;
                  end;
                  inc(i);
               end;
            end;
         end;
         While (i <= Length(txt)) and (Ord(Txt[i]) < 33) do begin
            if Txt[i] = #10 then
               lineSt := i+1;
            inc(i);
         end;
         st := i;
      end;
      i := 0;
      st := 0;
   end;
end;

function TNZCSSpellCheck.RemoveDupSpace(Memo: TCustomMemo; FirstDup: Integer): Boolean;
var
   St, SelSt: Integer;
begin
   result := True;
   Memo.SelStart := FirstDup;
   Memo.SelLength := 1;
   if not assigned(FrmSpellGrammer) then
      FrmSpellGrammer := TFrmSpellGrammer.Create(self);
   If FirstDup <= 5 then begin
      SelSt := 0;
      St := FirstDup;
   end
   else begin
      St := FirstDup - 5;
      SelSt := 5;
   end;
   FrmSpellGrammer.Caption := 'Duplicate Space Found:';
   FrmSpellGrammer.MemDisplay.Text := Copy(Memo.Lines.Text,St, 50);
   FrmSpellGrammer.MemDisplay.SelStart := SelSt;
   FrmSpellGrammer.MemDisplay.SelLength := 1;
   FrmSpellGrammer.Showmodal;
   if FrmSpellGrammer.CancelFlag = scCancel then
      result := false
   else begin
      if FrmSpellGrammer.CancelFlag = scChange then
         Memo.ClearSelection;
   end;
   Memo.SelLength := 0;
end;

function TNZCSSpellCheck.RepeativeWord(Memo: TCustomMemo; LastWordStart, LastWordEnd, StartWord, EndWord: Integer): Boolean;
var
   SelSt, st: Integer;
begin
   result := false;
   Memo.SelStart := LastWordStart-1;
   Memo.SelLength := LastWordEnd-LastWordStart + 1;
   if not assigned(FrmSpellGrammer) then
      FrmSpellGrammer := TFrmSpellGrammer.Create(self);
   If LastWordStart <= 5 then begin
      St := LastWordStart;
      SelSt := 0;
   end
   else begin
      St := LastWordStart - 5;
      SelSt := 5;
   end;
   FrmSpellGrammer.Caption := 'Repeative Word Found:';
   FrmSpellGrammer.MemDisplay.Text := Copy(Memo.Lines.Text,St, 50);
   FrmSpellGrammer.MemDisplay.SelStart := SelSt;
   FrmSpellGrammer.MemDisplay.SelLength := (LastWordEnd-LastWordStart);
   FrmSpellGrammer.Showmodal;
   if FrmSpellGrammer.CancelFlag = scChange then begin
      Memo.ClearSelection;
      result := True;
   end;
   {If Messagedlg('Repeative word found: ' + Copy(Memo.Lines.Text,LastWordStart, EndWord + 20) + CRLF +
      'Remove?',mtWarning,[MbYes, MbNo],0) = MRYes then begin
      Memo.ClearSelection;
      result := True;
   end;}
   Memo.SelLength := 0;
end;

function TNZCSSpellCheck.MisSpeltWord(Memo: TCustomMemo; Word: String; StartWord, EndWord: Integer): boolean;

Procedure ReplaceAll(FindTxt, ReplaceText: String);
var
   Found, frreplaceAll: Boolean;
begin
   Memo.SelStart := 0;
   Found := SearchMemo(Memo, Findtxt);
   frreplaceAll := True;
   If (not Found) then frreplaceAll := False;
   while (Found) do begin
       Memo.SelText := ReplaceText;
       If not frreplaceAll then exit;
       Found := SearchMemo(Memo, Findtxt);
   end;
   Memo.SelStart := StartWord + Length(ReplaceText);
end;


begin
   If not Assigned(FrmSpellCheck) then
      FrmSpellCheck := TFrmSpellCheck.Create(Self);
   FrmSpellCheck.NZCSSpellChecker := self;
   FrmSpellCheck.StartWord := StartWord;
   FrmSpellCheck.EndWord := EndWord;
   FrmSpellCheck.Memo := Memo;
   Memo.SelStart := StartWord-1;
   Memo.SelLength := EndWord-StartWord;
   FrmSpellCheck.EBWrong.Text := Word;
   if spAlwaysSuggest in SpellingOptions then begin
      GetSuggestions(Word, TStringList(FrmSpellCheck.lstSuggest.Items));
      If FrmSpellCheck.lstSuggest.Items.Count > 0 then
         FrmSpellCheck.EBChangeTo.Text := FrmSpellCheck.lstSuggest.Items[0];
   end;
   FrmSpellCheck.Showmodal;
   result := True;
   Case FrmSpellCheck.CancelFlag of
   scIgnoreAll: begin
                   IgnoreAllAdd(FrmSpellCheck.EBWrong.Text);
                   Memo.SelStart := StartWord + Length(FrmSpellCheck.EBWrong.Text);
                end;
   scIgnore: Memo.SelStart := StartWord + Length(FrmSpellCheck.EBWrong.Text);
   scCancel: result := False;
   scChange: begin
                Memo.SelText := FrmSpellCheck.EBChangeTo.Text;
                Memo.SelStart := StartWord + Length(FrmSpellCheck.EBChangeTo.Text);
             end;
   scChangeAll: begin
                   ReplaceAll(FrmSpellCheck.EBWrong.Text, FrmSpellCheck.EBChangeTo.Text);
                end;
   scAdd: CustomDictAdd(FrmSpellCheck.EBWrong.Text);
   scRestart: Begin
                 FIgnoreAll.Clear;
                 Memo.SelStart := 0;
              end;
   end;
   Memo.SelLength := 0;
end;

function TNZCSSpellCheck.SearchMemo(Memo: TCustomMemo;  const SearchString: String): Boolean;
var
   Buffer, P: PChar;
   Size: LongInt;

function SearchBuf(Buf: PChar; BufLen: Integer; SelStart, SelLength: Integer;
                   SearchString: String): PChar;
var
   SearchCount, I: Integer;
   C: Char;
   CharMap: array [Char] of Char;

function FindNextWordStart(var BufPtr: PChar): Boolean;
begin
   Result := False;
   while (SearchCount > 0) and (BufPtr^ in WordDelimiters) do begin
      Inc(BufPtr);
      Dec(SearchCount);
   end;
   Result := SearchCount >= 0;
end;

begin
   Result := nil;
   if BufLen <= 0 then Exit;
   Inc(SelStart, SelLength);  { start search past end of selection }
   SearchCount := succ(BufLen - SelStart - Length(SearchString));
   if SearchCount < 0 then Exit;
   if Longint(SelStart) + SearchCount > BufLen then Exit;
   if (SelStart < 0) or (SelStart > BufLen) then Exit;
   Result := @Buf[SelStart];
   for C := Low(CharMap) to High(CharMap) do
      CharMap[C] := C;
   while SearchCount >= 0 do begin
      if not FindNextWordStart(Result) then Break;
      I := 0;
      while (CharMap[Result[I]] = SearchString[I + 1]) do begin
         Inc(I);
         if I >= Length(SearchString) then begin
            if (SearchCount = 0) or (Result[I] in WordDelimiters) then
               Exit;
            Break;
         end;
      end;
      Inc(Result);
      Dec(SearchCount);
   end;
   Result := nil;
end;

begin
   Result := False;
   if (Length(SearchString) = 0) then Exit;
   Size := Memo.GetTextLen;
   if (Size = 0) then Exit;
   Buffer := StrAlloc(Size + 1);
   try
      Memo.GetTextBuf(Buffer, Size + 1);
      P := SearchBuf(Buffer, Size, Memo.SelStart, Memo.SelLength,
      SearchString);
      if P <> nil then begin
         Memo.SelStart := P - Buffer;
         Memo.SelLength := Length(SearchString);
         Result := True;
      end;
   finally
      StrDispose(Buffer);
   end;
end;

function TNZCSSpellCheck.FindCustomDic: String;
const
   AppDataPath = 'Microsoft\Proof\Custom.Dic';
   AppDataKey = 'Software\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders';
   Win95Cust = 'SOFTWARE\Microsoft\Shared Tools\Proofing Tools\Custom Dictionaries';
   Win95Def = 'C:\Program Files\Common Files\Microsoft Shared\Proof\';
var
   myRegistry : TRegistry;
   AddDataDir: String;
begin
   // If Win98 / NT / 2000 then get the AppData directory
   myRegistry := TRegistry.Create;
   myRegistry.RootKey := HKEY_CURRENT_USER;
   If myRegistry.OpenKey(AppDataKey, false) then begin
      AddDataDir := myRegistry.ReadString('AppData');
      If (AddDataDir <> '') and (AddDataDir[Length(AddDataDir)] <> '\') then
         AddDataDir := AddDataDir + '\';
      MyRegistry.CloseKey;
   end;
   If (AddDataDir <> '') and FileExists(AddDataDir + AppDataPath) then
      result := AddDataDir + AppDataPath
   else begin
      // If Win95 the it will probably be in the Office directory //
      myRegistry.RootKey := HKEY_LOCAL_MACHINE;
      If myRegistry.OpenKey(Win95Cust, False) then begin
         result := myRegistry.ReadString('1');
         If (result <> '') and (ExtractFilePath(result) = '') then
            result := '';
         MyRegistry.CloseKey;
      end;
      If result = '' then begin
         // OK cannot find one so lets create one in a reasonably place //
         If AddDataDir <> '' then begin
            ForceDirectories(ExtractFilePath(AddDataDir + AppDataPath));
            result := AddDataDir + AppDataPath;
         end
         else begin
            ForceDirectories(Win95Def);
            result := Win95Def + 'custom.dic';
         end;
      end;
   end;
   MyRegistry.Free;
end;

Procedure TNZCSSpellCheck.LoadSpellOptions;
var
   myRegistry : TRegistry;
begin
   myRegistry := TRegistry.Create;
   myRegistry.RootKey := HKEY_CURRENT_USER;
   If myRegistry.OpenKey(SpellRegPath + '\Options', false) then begin
      If myRegistry.ReadString('IgnoreNumbers') <> '' then begin
         SpellingOptions := [];
         If myRegistry.ReadString('IgnoreNumbers') = '1' then
            SpellingOptions := SpellingOptions + [spIgnoreNumbers];
         If myRegistry.ReadString('IgnoreAllUpperCase') = '1' then
            SpellingOptions := SpellingOptions + [spIgnoreAllUpperCase];
         If myRegistry.ReadString('IgnoreInternet') = '1' then
            SpellingOptions := SpellingOptions + [spIgnoreInternet];
         If myRegistry.ReadString('IgnoreFiles') = '1' then
            SpellingOptions := SpellingOptions + [spIgnoreFiles];
         If myRegistry.ReadString('IgnoreBrackets') = '1' then
            SpellingOptions := SpellingOptions + [spIgnoreBrackets];
         If myRegistry.ReadString('MainDicOnly') = '1' then
            SpellingOptions := SpellingOptions + [spMainDicOnly];
         If myRegistry.ReadString('DuplicateSpaces') = '1' then
            SpellingOptions := SpellingOptions + [spDuplicateSpaces];
         If myRegistry.ReadString('DuplicateWords') = '1' then
            SpellingOptions := SpellingOptions + [spDuplicateWords];
         If myRegistry.ReadString('AlwaysSuggest') = '1' then
            SpellingOptions := SpellingOptions + [spAlwaysSuggest];
         If FileExists(myRegistry.ReadString('CustomDictionary')) then
            CustomDictionary := myRegistry.ReadString('CustomDictionary');
      end;
      myRegistry.CloseKey;
   end;
   If myRegistry.OpenKey(SpellRegPath, True) then begin
      If (myRegistry.ReadString('Dictionary') <> '') and
         FileExists(myRegistry.ReadString('Dictionary')) then
         Dictionary := myRegistry.ReadString('Dictionary');
   end;
   myRegistry.Free;
end;

Procedure TNZCSSpellCheck.SaveSpellOptions;
var
   myRegistry : TRegistry;
begin
   myRegistry := TRegistry.Create;
   myRegistry.RootKey := HKEY_CURRENT_USER;
   If myRegistry.OpenKey(SpellRegPath + '\Options', True) then begin
      If spIgnoreNumbers in SpellingOptions then
         myRegistry.WriteString('IgnoreNumbers','1')
      else
         myRegistry.WriteString('IgnoreNumbers','0');
      If spIgnoreAllUpperCase in SpellingOptions then
         myRegistry.WriteString('IgnoreAllUpperCase','1')
      else
         myRegistry.WriteString('IgnoreAllUpperCase','0');
      If spIgnoreInternet in SpellingOptions then
         myRegistry.WriteString('IgnoreInternet','1')
      else
         myRegistry.WriteString('IgnoreInternet','0');
      If spIgnoreFiles in SpellingOptions then
         myRegistry.WriteString('IgnoreFiles','1')
      else
         myRegistry.WriteString('IgnoreFiles','0');
      If spIgnoreBrackets in SpellingOptions then
         myRegistry.WriteString('IgnoreBrackets','1')
      else
         myRegistry.WriteString('IgnoreBrackets','0');
      If spMainDicOnly in SpellingOptions then
         myRegistry.WriteString('MainDicOnly','1')
      else
         myRegistry.WriteString('MainDicOnly','0');
      If spDuplicateSpaces in SpellingOptions then
         myRegistry.WriteString('DuplicateSpaces','1')
      else
         myRegistry.WriteString('DuplicateSpaces','0');
      If spDuplicateWords in SpellingOptions then
         myRegistry.WriteString('DuplicateWords','1')
      else
         myRegistry.WriteString('DuplicateWords','0');
      If spAlwaysSuggest in SpellingOptions then
         myRegistry.WriteString('AlwaysSuggest','1')
      else
         myRegistry.WriteString('AlwaysSuggest','0');
      If (CustomDictionary <> '') and FileExists(myRegistry.ReadString('CustomDictionary')) then
         myRegistry.WriteString('CustomDictionary',CustomDictionary);
      myRegistry.CloseKey;
   end;
   If myRegistry.OpenKey(SpellRegPath, True) then begin
      myRegistry.WriteString('Dictionary',Dictionary);
      myRegistry.CloseKey;
   end;
   myRegistry.Free;
end;


Function TNZCSSpellCheck.WordAtCursor(Memo: TMemo): String;
var
   l_CurrentLine: String;
   l_wordend,i: Smallint;
   Pt: TPoint;
   Index, Line, cpMin, cpMax: Longint;
   wordstart: Integer;

function IsValIdent(const Ident: Char): Boolean;
const
  Ascii = ['A'..'Z', 'a'..'z', '_', '.','0'..'9'];
begin
  Result := Ident in Ascii;
end;

begin
   Result := '';
   Memo.Perform(em_GetSel, WPARAM(@cpMin), LPARAM(@cpMax));
   Index:=Abs(cpMin);
   Line:=Memo.Perform(em_LineIndex, -1, 0);
   Pt.X:=Abs(Index-Line+1);
   Pt.Y:=Memo.Perform(em_LineFromChar, Index, 0);
   If Pt.Y >= 0 then begin
      l_CurrentLine := Memo.Lines[Pt.Y];
      If l_CurrentLine = '' then
         Exit;
      If Memo.SelText <> '' then
         i := Pt.X + Memo.SelLength
      else
         i := Pt.X;
      if i > length(l_CurrentLine) then
         i := length(l_CurrentLine);
      If i < 1 then
         i := 1;
      While (i > 0) and not (IsValIdent(l_CurrentLine[i])) do
         dec(i);
      If (i > 0) and IsValIdent(l_CurrentLine[i]) or (i = length(l_CurrentLine)) then begin
         If (not IsValIdent(l_CurrentLine[i])) and (i = length(l_CurrentLine)) then
            dec(i);
         If i < 1 then
            i := 1;
          // We are part way though a word so dec to the start //
         While (i > 0) and (IsValIdent(l_CurrentLine[i])) do
            dec(i);
      end;
      While (i <= length(l_CurrentLine)) and (not IsValIdent(l_CurrentLine[i])) do
         inc(i);
      wordstart := i;
      While (i <= length(l_CurrentLine)) and (IsValIdent(l_CurrentLine[i])) do
         inc(i);
      l_wordend := i - wordstart;
      result := Copy(l_CurrentLine,wordstart,l_wordend);
      If result <> '' then begin
         memo.SelStart := Line + wordstart-1;
         memo.SelLength:= l_wordend;
      end;
   end;
end;



end.



